#include "SubjectIf.h"
#include "ObserverIf.h"

#include <iostream>
#include <mutex>
#include <set>

namespace
{

class Subject : public SubjectIf
{
public:
    Subject();

    std::unique_ptr<RegistrationTokenIf> registerObserver(ObserverIf& observer) override;

    void actAsIfStateWouldChange();

    void unRegisterObserver(ObserverIf& observer);

private:

    void notify(const std::lock_guard<std::mutex>&);

    std::set<ObserverIf*> mObservers;
    std::mutex mMutex;
};

class RegistrationToken : public RegistrationTokenIf
{
public:
    RegistrationToken(
            Subject& subject,
            ObserverIf& observer) :
        mSubject{subject},
        mObserver{observer}
    {
    }

    ~RegistrationToken()
    {
        mSubject.unRegisterObserver(mObserver);
    }

private:
    Subject& mSubject;
    ObserverIf& mObserver;
};


Subject::Subject() :
    mObservers{},
    mMutex{}
{
}

std::unique_ptr<RegistrationTokenIf> Subject::registerObserver(ObserverIf& observer)
{
    std::lock_guard lock{mMutex};
    mObservers.insert(&observer);
    return std::unique_ptr<RegistrationTokenIf>{new RegistrationToken(*this, observer)};
}

void Subject::actAsIfStateWouldChange()
{
    std::lock_guard lock{mMutex};
    notify(lock);
}

void Subject::unRegisterObserver(ObserverIf& observer)
{
    std::lock_guard lock{mMutex};
    mObservers.erase(&observer);
}

void Subject::notify(const std::lock_guard<std::mutex>&)
{
    for(auto observer : mObservers)
    {
        observer->update();
    }
}


class Observer : public ObserverIf
{
public:
    explicit Observer(const std::string& name) :
        mName{name}
    {
    }

    void update() override
    {
        std::cout << mName << std::endl;
    }
private:
    std::string mName;
};

} // namespace

int main()
{
    Subject sub;

    {
        Observer foo{"foo"};
        Observer bar{"bar"};
        Observer real{"real"};
        Observer nope{"nope"};

        auto tokenFoo = sub.registerObserver(foo);
        auto tokenBar = sub.registerObserver(bar);
        auto tokenReal = sub.registerObserver(real);
        auto tokenNope = sub.registerObserver(nope);

        tokenNope.reset();

        sub.actAsIfStateWouldChange();
    }
    std::cout << "once more " << std::endl;
    sub.actAsIfStateWouldChange();
}
