#include "SubjectIf.h"

#include <iostream>
#include <mutex>
#include <map>
#include <cstdint>

namespace
{

class Subject : public SubjectIf
{
public:
    Subject();

    std::unique_ptr<RegistrationTokenIf> registerObserver(const Observer& observer) override;

    void actAsIfStateWouldChange();

    void unRegisterObserver(uint64_t id);

private:

    void notify(const std::lock_guard<std::mutex>&);

    std::map<int, Observer> mObservers;
    uint64_t mId; /// will dry up some day
    std::mutex mMutex;
};

class RegistrationToken : public RegistrationTokenIf
{
public:
    RegistrationToken(
            Subject& subject,
            uint64_t id) :
        mSubject{subject},
        mId{id}
    {
    }

    ~RegistrationToken()
    {
        mSubject.unRegisterObserver(mId);
    }

private:
    Subject& mSubject;
    uint64_t mId;
};


Subject::Subject() :
    mObservers{},
    mId{},
    mMutex{}
{
}

std::unique_ptr<RegistrationTokenIf> Subject::registerObserver(const Observer& observer)
{
    std::lock_guard lock{mMutex};
    mObservers[mId] = observer;
    return std::unique_ptr<RegistrationTokenIf>{new RegistrationToken(*this, mId++)};
}

void Subject::actAsIfStateWouldChange()
{
    std::lock_guard lock{mMutex};
    notify(lock);
}

void Subject::unRegisterObserver(uint64_t id)
{
    std::lock_guard lock{mMutex};
    mObservers.erase(id);
}

void Subject::notify(const std::lock_guard<std::mutex>&)
{
    for(auto observer : mObservers)
    {
        observer.second();
    }
}

class MyObserver
{
public:
    explicit MyObserver(const std::string& name) :
        mName{name}
    {
    }

    void update()
    {
        std::cout << "my : " << mName << std::endl;
    }
private:
    std::string mName;
};

class YourObserver
{
public:
    explicit YourObserver(const std::string& name) :
        mName{name}
    {
    }

    void operator()()
    {
        std::cout << "your : " << mName << std::endl;
    }
private:
    std::string mName;
};

void foo()
{
    std::cout << "foo() observing" << std::endl;
}

void nope()
{
    std::cout << "nope() observing" << std::endl;
}

} // namespace

int main()
{
    Subject sub;

    {
        MyObserver bar{"bar"};
        YourObserver real{"real"};

        auto tokenFoo = sub.registerObserver(foo);
        auto tokenBar = sub.registerObserver(std::bind(&MyObserver::update, &bar));
        auto tokenReal = sub.registerObserver(real);
        auto tokenNope = sub.registerObserver(nope);
        auto tokenLambda = sub.registerObserver([](){std::cout << "lambda" << std::endl;});

        tokenNope.reset();

        sub.actAsIfStateWouldChange();
    }
    std::cout << "once more " << std::endl;
    sub.actAsIfStateWouldChange();
}
