#include <boost/signals2.hpp>

#include <iostream>
#include <thread>


/// OBSERVER PATTERN
///
///     => subject      ---> signal
///
///     => observer     ---> slot
///
///     => notification (signal) going from subject to observer(s)

/// observer registers on the subject
///
///     => observer CONNECTS on the subject ===> CONNECTION is created
///
///     NOT !!!!!!!  RAII : keep track of the connection in order to unregister
///
///     => registers the "CALLBACK", the entity to call (aka signal)

/// make the connection RAII ===> boost::signals2::scoped_connection


namespace
{

class Subject
{
public:
    using Signal = boost::signals2::signal<void ()>;
    using Slot = Signal::slot_type;

    boost::signals2::scoped_connection registerObserver(
            const Slot& handler)
    {
        return mSignalHandlers.connect(handler);
    }

    void notify()
    {
        mSignalHandlers();
    }

private:

    Signal mSignalHandlers; /// aka the observers
};

void foo()
{
    std::cout << "foo" << std::endl;
}

void bar()
{
    std::cout << "bar" << std::endl;
}

void real()
{
    std::cout << "real" << std::endl;
}

void nope()
{
    std::cout << "nope" << std::endl;
}

class Yoyo
{
public:
    Yoyo() = default;

    void doSomething()
    {
        std::cout << "Yoyo" << std::endl;
    }
};

} // namespace

int main()
{
    Subject sub;

    {
        auto connectionFoo{sub.registerObserver(foo)};
        auto connectionBar{sub.registerObserver(bar)};
        auto connectionReal{sub.registerObserver(real)};
        auto connectionNope{sub.registerObserver(nope)};
        Yoyo yoyo;
        auto connectionYoyo{sub.registerObserver(Subject::Slot(&Yoyo::doSomething, yoyo))};
        auto connectionLambda{sub.registerObserver([](){std::cout << "lambda" << std::endl;})};

        connectionNope.disconnect();

        sub.notify();
    }

    std::cout << "once more " << std::endl;
    sub.notify();
}
